@@ -2,11 +2,13 @@
 
 import asyncio
 import enum
-import functools
 import io
+import os
 
-from shellous.log import LOGGER
-from shellous.util import decode, log_method
+from shellous.log import log_method
+from shellous.util import decode
+
+_DETAILED_LOGGING = True
 
 
 class Redirect(enum.IntEnum):
@@ -22,8 +24,15 @@ class Redirect(enum.IntEnum):
         return self in {Redirect.CAPTURE, Redirect.INHERIT}
 
 
-@log_method(True)
+# Used in Command and Pipeline to implement operator overloading.
+STDIN_TYPES = (str, bytes, os.PathLike, bytearray, io.IOBase, int, Redirect)
+STDOUT_TYPES = (str, bytes, os.PathLike, bytearray, io.IOBase, int, Redirect)
+STDOUT_APPEND_TYPES = (str, bytes, os.PathLike)
+
+
+@log_method(_DETAILED_LOGGING)
 async def write_stream(input_bytes, stream):
+    "Write input_bytes to stream."
     try:
         if input_bytes:
             stream.write(input_bytes)
@@ -32,12 +41,15 @@ async def write_stream(input_bytes, stream):
         # Catch these errors and quietly stop.
         # See "_feed_stdin" in /3.9/Lib/asyncio/subprocess.py
         pass
-    finally:
-        stream.close()
+
+    # If `stream.drain()`` is cancelled, we do NOT close the stream here.
+    # See https://bugs.python.org/issue45074
+    stream.close()
 
 
-@log_method(True)
+@log_method(_DETAILED_LOGGING)
 async def copy_stringio(source, dest, encoding):
+    "Copy bytes from source stream to dest StringIO."
     # Collect partial reads into a BytesIO.
     buf = io.BytesIO()
     try:
@@ -52,8 +64,9 @@ async def copy_stringio(source, dest, encoding):
         dest.write(decode(buf.getvalue(), encoding))
 
 
-@log_method(True)
+@log_method(_DETAILED_LOGGING)
 async def copy_bytesio(source, dest):
+    "Copy bytes from source stream to dest BytesIO."
     # Collect partial reads into a BytesIO.
     while True:
         data = await source.read(1024)
@@ -62,8 +75,9 @@ async def copy_bytesio(source, dest):
         dest.write(data)
 
 
-@log_method(True)
+@log_method(_DETAILED_LOGGING)
 async def copy_bytearray(source, dest):
+    "Copy bytes from source stream to dest bytearray."
     # Collect partial reads into a bytearray.
     while True:
         data = await source.read(1024)
