@@ -6,17 +6,21 @@ import os
 import sys
 
 import shellous.redirect as redir
-from shellous.log import LOGGER
+from shellous.harvest import harvest, harvest_results
+from shellous.log import LOGGER, log_method
 from shellous.redirect import Redirect
 from shellous.result import Result, make_result
-from shellous.util import decode, gather_collect, log_method, platform_info
+from shellous.util import close_fds, decode
 
+_NORMAL_LOGGING = True
 _DETAILED_LOGGING = True
 
+_KILL_TIMEOUT = 3.0
+_CLOSE_TIMEOUT = 0.25
 
-def _exc():
-    "Return the current exception value. Useful in logging."
-    return sys.exc_info()[1]
+
+def _is_cancelled(ex):
+    return isinstance(ex, asyncio.CancelledError)
 
 
 class _RunOptions:
@@ -42,7 +46,7 @@ class _RunOptions:
 
     def close_fds(self):
         "Close all open file descriptors in `open_fds`."
-        _close_fds(self.open_fds)
+        close_fds(self.open_fds)
 
     def __enter__(self):
         "Set up I/O redirections."
@@ -57,7 +61,7 @@ class _RunOptions:
         "Make sure those file descriptors are cleaned up."
         self.close_fds()
         if exc_value:
-            LOGGER.warning(
+            LOGGER.info(
                 "_RunOptions.exit %r exc_value=%r", self.command.name, exc_value
             )
 
@@ -162,10 +166,9 @@ class Runner:
     """Runner is an asynchronous context manager that runs a command.
 
     ```
-    runner = Runner(cmd)
-    async with runner as (stdin, stdout, stderr):
-        # process stdin, stdout, stderr (if not None)
-    result = runner.result()
+    async with cmd.run() as run:
+        # process run.stdin, run.stdout, run.stderr (if not None)
+    result = run.result()
     ```
     """
 
@@ -175,6 +178,8 @@ class Runner:
         self.proc = None
         self.tasks = []
         self.stdin = None
+        self.stdout = None
+        self.stderr = None
 
     @property
     def name(self):
@@ -200,7 +205,7 @@ class Runner:
             self.options.encoding,
         )
 
-        return make_result(self.command, result)
+        return make_result(self.command, result, self.cancelled)
 
     def add_task(self, coro, tag=None):
         "Add a background task."
@@ -217,7 +222,7 @@ class Runner:
         assert self.proc
 
         try:
-            await gather_collect(*self.tasks, trustee=self)
+            await harvest(*self.tasks, trustee=self)
 
         except asyncio.CancelledError:
             LOGGER.info("Runner.wait cancelled %r", self)
@@ -234,19 +239,21 @@ class Runner:
         cancel_signal = self.command.options.cancel_signal
 
         try:
+            done = self.proc.returncode is not None
+
             # If not already done, send cancel signal.
-            if self.proc.returncode is None:
+            if not done:
                 self._send_signal(cancel_signal)
 
             if self.tasks:
-                await gather_collect(*self.tasks, timeout=cancel_timeout, trustee=self)
-            else:
-                await gather_collect(
-                    self.proc.wait(), timeout=cancel_timeout, trustee=self
-                )
+                await harvest(*self.tasks, timeout=cancel_timeout, trustee=self)
+            elif not done:
+                await harvest(self.proc.wait(), timeout=cancel_timeout, trustee=self)
 
         except (asyncio.CancelledError, asyncio.TimeoutError) as ex:
             LOGGER.warning("Runner.kill %r (ex)=%r", self, ex)
+            if _is_cancelled(ex):
+                self.cancelled = True
             await self._kill_wait()
 
         except Exception as ex:
@@ -272,18 +279,24 @@ class Runner:
         if self.proc.returncode is not None:
             return
 
-        self.proc.kill()
-        await self.proc.wait()  # FIXME: needs timeout...
+        try:
+            self.proc.kill()
+            await harvest(self.proc.wait(), timeout=_KILL_TIMEOUT, trustee=self)
+        except asyncio.TimeoutError as ex:
+            LOGGER.error("%r failed to kill process %r", self, self.proc)
+            raise RuntimeError(f"Unable to kill process {self.proc!r}") from ex
 
+    @log_method(_NORMAL_LOGGING, _info=True)
     async def __aenter__(self):
         "Set up redirections and launch subprocess."
-        LOGGER.info("Runner entering %r (%s)", self, platform_info())
         try:
             return await self._start()
-
         finally:
-            LOGGER.info("Runner entered %r ex=%r", self, _exc())
+            if self.cancelled and self.command.options.incomplete_result:
+                # Raises ResultError instead of CancelledError.
+                self.result()
 
+    @log_method(_DETAILED_LOGGING)
     async def _start(self):
         "Set up redirections and launch subprocess."
         assert self.proc is None
@@ -304,9 +317,6 @@ class Runner:
             stdout = self.proc.stdout
             stderr = self.proc.stderr
 
-            # Keep track of stdin so we can close it later.
-            self.stdin = stdin
-
             if stderr is not None:
                 error = opts.command.options.error
                 stderr = self._setup_output_sink(stderr, error, opts.encoding, "stderr")
@@ -323,13 +333,20 @@ class Runner:
                     stdin = None
 
         except (Exception, asyncio.CancelledError) as ex:
-            LOGGER.warning("Runner.start %r ex=%r", self, ex)
-            self.cancelled = isinstance(ex, asyncio.CancelledError)
+            LOGGER.info("Runner.start %r ex=%r", self, ex)
+            if _is_cancelled(ex):
+                self.cancelled = True
             if self.proc:
                 await self._kill()
             raise
 
-        return (stdin, stdout, stderr)
+        # Make final streams available. These may be different from `self.proc`
+        # versions.
+        self.stdin = stdin
+        self.stdout = stdout
+        self.stderr = stderr
+
+        return self
 
     def _setup_output_sink(self, stream, sink, encoding, tag):
         "Set up a task to write to custom output sink."
@@ -344,32 +361,26 @@ class Runner:
             stream = None
         return stream
 
+    @log_method(_NORMAL_LOGGING, exc_value=2)
     async def __aexit__(self, _exc_type, exc_value, _exc_tb):
         "Wait for process to exit and handle cancellation."
-
-        LOGGER.info("Runner exiting %r exc_value=%r", self, exc_value)
         suppress = False
         try:
             suppress = await self._finish(exc_value)
         except asyncio.CancelledError:
-            LOGGER.warning("Runner cancelled inside _finish %r", self)
-        finally:
-            LOGGER.info(
-                "Runner exited %r ex=%r exc_value=%r suppress=%r",
-                self,
-                _exc(),
-                exc_value,
-                suppress,
-            )
+            LOGGER.info("Runner cancelled inside _finish %r", self)
+            self.cancelled = True
         return suppress
 
+    @log_method(_DETAILED_LOGGING)
     async def _finish(self, exc_value):
         "Finish the run. Return True only if `exc_value` should be suppressed."
         assert self.proc
 
         try:
             if exc_value is not None:
-                self.cancelled = isinstance(exc_value, asyncio.CancelledError)
+                if _is_cancelled(exc_value):
+                    self.cancelled = True
                 await self._kill()
                 return self.cancelled
 
@@ -383,51 +394,82 @@ class Runner:
     async def _close(self):
         "Make sure that our resources are properly closed."
         assert self.proc
-        assert self.proc.returncode is not None
+
+        # _close can be called when unwinding exceptions. We need to handle
+        # the case that the process has not exited yet. Remember to close the
+        # transport.
+        if self.proc.returncode is None:
+            LOGGER.critical("Runner._close process still running %r", self.proc)
+            self.proc._transport.close()
+            return
 
         try:
             # Make sure the transport is closed (for asyncio and uvloop).
             self.proc._transport.close()
 
-            # Make sure that stdin is properly closed. `wait_closed` will raise
-            # a BrokenPipeError if not all input was properly written.
-            if self.stdin is not None:
-                self.stdin.close()
-                await gather_collect(
-                    self.stdin.wait_closed(), timeout=0.25, trustee=self
+            # Make sure that original stdin is properly closed. `wait_closed`
+            # will raise a BrokenPipeError if not all input was properly written.
+            if self.proc.stdin is not None:
+                self.proc.stdin.close()
+                await harvest(
+                    self.proc.stdin.wait_closed(),
+                    timeout=_CLOSE_TIMEOUT,
+                    trustee=self,
                 )
 
         except asyncio.TimeoutError:
-            LOGGER.error("Runner._close %r timeout", self)
+            LOGGER.critical("Runner._close %r timeout stdin=%r", self, self.proc.stdin)
 
     def __repr__(self):
         "Return string representation of Runner."
+        cancelled = " cancelled" if self.cancelled else ""
         if self.proc:
-            return f"<Runner {self.name!r} pid={self.proc.pid} exit_code={self.proc.returncode}>"
-        return f"<Runner {self.name!r} pid=None>"
+            procinfo = f" pid={self.proc.pid} exit_code={self.proc.returncode}"
+        else:
+            procinfo = " pid=None"
+        return f"<Runner {self.name!r}{cancelled}{procinfo}>"
+
+    async def _readlines(self):
+        "Iterate over lines in stdout/stderr"
+        if self.stdin or (self.stdout and self.stderr):
+            raise RuntimeError("multiple capture not supported in iterator")
+
+        encoding = self.options.encoding
+        stream = self.stdout or self.stderr
+        if stream:
+            async for line in stream:
+                yield decode(line, encoding)
+
+    def __aiter__(self):
+        "Return asynchronous iterator over stdout/stderr."
+        return self._readlines()
 
 
-class PipeRunner:
+class PipeRunner:  # pylint: disable=too-many-instance-attributes
     """PipeRunner is an asynchronous context manager that runs a pipeline.
 
     ```
-    runner = PipeRunner(pipe)
-    async with runner as (stdin, stdout, stderr):
-        # process stdin, stdout, stderr (if not None)
-    result = runner.result()
+    async with pipe.run() as run:
+        # process run.stdin, run.stdout, run.stderr (if not None)
+    result = run.result()
     ```
     """
 
-    def __init__(self, pipe, *, capturing=False):
+    def __init__(self, pipe, *, capturing):
         """`capturing=True` indicates we are within an `async with` block and
         client needs to access `stdin` and `stderr` streams.
         """
+        assert len(pipe.commands) > 1
+
         self.pipe = pipe
         self.cancelled = False
         self.tasks = None
         self.results = None
         self.capturing = capturing
         self.encoding = pipe.commands[-1].options.encoding
+        self.stdin = None
+        self.stdout = None
+        self.stderr = None
 
     @property
     def name(self):
@@ -436,7 +478,7 @@ class PipeRunner:
 
     def result(self):
         "Return `Result` object for PipeRunner."
-        return make_result(self.pipe, self.results)
+        return make_result(self.pipe, self.results, self.cancelled)
 
     @log_method(_DETAILED_LOGGING)
     async def _wait(self, *, kill=False):
@@ -449,45 +491,43 @@ class PipeRunner:
             for task in self.tasks:
                 task.cancel()
 
-        self.results = await gather_collect(
-            *self.tasks, return_exceptions=True, trustee=self
-        )
+        cancelled, self.results = await harvest_results(*self.tasks, trustee=self)
+        if cancelled:
+            self.cancelled = True
 
+    @log_method(_NORMAL_LOGGING)
     async def __aenter__(self):
         "Set up redirections and launch pipeline."
-        LOGGER.info("PipeRunner entering %r", self)
         try:
             return await self._start()
         except (Exception, asyncio.CancelledError) as ex:
             LOGGER.warning("PipeRunner enter %r ex=%r", self, ex)
-            await self.wait(kill=True)  # FIXME
+            if _is_cancelled(ex):
+                self.cancelled = True
+            await self._wait(kill=True)  # FIXME
             raise
-        finally:
-            LOGGER.info("PipeRunner entered %r ex=%r", self, _exc())
 
+    @log_method(_NORMAL_LOGGING, exc_value=2)
     async def __aexit__(self, _exc_type, exc_value, _exc_tb):
         "Wait for pipeline to exit and handle cancellation."
-        LOGGER.info("PipeRunner exiting %r exc_value=%r", self, exc_value)
+        suppress = False
         try:
-            if exc_value is not None:
-                return await self._cleanup(exc_value)
-            await self._wait()
-        finally:
-            LOGGER.info(
-                "PipeRunner exited %r exc_value=%r ex=%r",
-                self,
-                exc_value,
-                _exc(),
-            )
+            suppress = await self._finish(exc_value)
+        except asyncio.CancelledError:
+            LOGGER.warning("PipeRunner cancelled inside _finish %r", self)
+            self.cancelled = True
+        return suppress
 
     @log_method(_DETAILED_LOGGING)
-    async def _cleanup(self, exc_value):
-        "Clean up when there is an exception."
+    async def _finish(self, exc_value):
+        "Wait for pipeline to exit and handle cancellation."
+        if exc_value is not None:
+            if _is_cancelled(exc_value):
+                self.cancelled = True
+            await self._wait(kill=True)
+            return self.cancelled
 
-        self.cancelled = isinstance(exc_value, asyncio.CancelledError)
-        await self._wait(kill=True)
-
-        return self.cancelled
+        await self._wait()
 
     @log_method(_DETAILED_LOGGING)
     async def _start(self):
@@ -506,11 +546,15 @@ class PipeRunner:
             else:
                 self.tasks = [cmd.task() for cmd in cmds]
 
-            return (stdin, stdout, stderr)
+            self.stdin = stdin
+            self.stdout = stdout
+            self.stderr = stderr
 
-        except BaseException:
+            return self
+
+        except BaseException:  # pylint: disable=broad-except
             # Clean up after any exception *including* CancelledError.
-            _close_fds(open_fds)
+            close_fds(open_fds)
 
     def _setup_pipeline(self, open_fds):
         """Return the pipeline stitched together with pipe fd's.
@@ -529,7 +573,7 @@ class PipeRunner:
             cmds[i + 1] = cmds[i + 1].stdin(read_fd, close=True)
 
         for i in range(cmd_count):
-            cmds[i] = cmds[i].set(return_result=True)
+            cmds[i] = cmds[i].set(return_result=True, incomplete_result=True)
 
         return cmds
 
@@ -540,109 +584,75 @@ class PipeRunner:
         loop = asyncio.get_event_loop()
         first_fut = loop.create_future()
         last_fut = loop.create_future()
-        first_task = cmds[0].task(_streams_future=first_fut)
-        last_task = cmds[-1].task(_streams_future=last_fut)
+        first_task = cmds[0].task(_run_future=first_fut)
+        last_task = cmds[-1].task(_run_future=last_fut)
 
         middle_tasks = [cmd.task() for cmd in cmds[1:-1]]
         self.tasks = [first_task] + middle_tasks + [last_task]
 
         # When capturing, we need the first and last commands in the
         # pipe to signal when they are ready.
-        first_ready, last_ready = await gather_collect(
-            first_fut, last_fut, trustee=self
+        first_ready, last_ready = await asyncio.gather(first_fut, last_fut)
+
+        stdin, stdout, stderr = (
+            first_ready.stdin,
+            last_ready.stdout,
+            last_ready.stderr,
         )
-        stdin, stdout, stderr = (first_ready[0], last_ready[1], last_ready[2])
 
         return (stdin, stdout, stderr)
 
     def __repr__(self):
         "Return string representation of PipeRunner."
+        cancelled_info = ""
+        if self.cancelled:
+            cancelled_info = " cancelled"
         result_info = ""
         if self.results:
             result_info = f" results={self.results!r}"
-        return f"<PipeRunner {self.name!r}{result_info}>"
+        return f"<PipeRunner {self.name!r}{cancelled_info}{result_info}>"
 
+    async def _readlines(self):
+        "Iterate over lines in stdout/stderr"
+        if self.stdin or (self.stdout and self.stderr):
+            raise RuntimeError("multiple capture not supported in iterator")
 
-async def run(command, *, _streams_future=None):
-    "Run a command."
-    if not _streams_future and command.multiple_capture:
-        raise ValueError("multiple capture requires 'async with'")
-
-    output_bytes = None
-    runner = Runner(command)
-
-    try:
-        async with runner as (stdin, stdout, stderr):
-            if _streams_future is not None:
-                # Return streams to caller in another task.
-                _streams_future.set_result((stdin, stdout, stderr))
-
-            else:
-                # Read the output here and return it.
-                stream = stdout or stderr
-                if stream:
-                    output_bytes = await stream.read()
-
-    except asyncio.CancelledError:
-        LOGGER.info("run %r cancelled inside enter", command.name)
-
-    return runner.result(output_bytes)
-
-
-async def run_iter(command):
-    "Run a command and iterate over its output lines."
-    if command.multiple_capture:
-        raise ValueError("multiple capture requires 'async with'")
-
-    runner = Runner(command)
-    async with runner as (_stdin, stdout, stderr):
-        encoding = runner.options.encoding
-        stream = stdout or stderr
+        encoding = self.encoding
+        stream = self.stdout or self.stderr
         if stream:
             async for line in stream:
                 yield decode(line, encoding)
 
-    runner.result()  # No return value; raises exception if needed
+    def __aiter__(self):
+        "Return asynchronous iterator over stdout/stderr."
+        return self._readlines()
+
+
+async def run_cmd(command, *, _run_future=None):
+    "Run a command."
+    if not _run_future and command.multiple_capture:
+        raise ValueError("multiple capture requires 'async with'")
+
+    output_bytes = None
+
+    async with command.run() as run:
+        if _run_future is not None:
+            # Return streams to caller in another task.
+            _run_future.set_result(run)
+
+        else:
+            # Read the output here and return it.
+            stream = run.stdout or run.stderr
+            if stream:
+                output_bytes = await stream.read()
+
+    return run.result(output_bytes)
 
 
 async def run_pipe(pipe):
     "Run a pipeline"
 
-    cmd_count = len(pipe.commands)
-    if cmd_count == 1:
-        return await run(pipe.commands[0])
-
-    runner = PipeRunner(pipe)
-    async with runner:
+    run = PipeRunner(pipe, capturing=False)
+    async with run:
         pass
-    return runner.result()
-
-
-async def run_pipe_iter(pipe):
-    "Run a pipeline and iterate over its output lines."
-    runner = PipeRunner(pipe, capturing=True)
-    async with runner as (stdin, stdout, stderr):
-        assert stdin is None
-        assert stderr is None
-
-        encoding = runner.encoding
-        async for line in stdout:
-            yield decode(line, encoding)
-
-    runner.result()  # No return value; raises exception if needed
-
-
-def _close_fds(open_fds):
-    "Close open file descriptors or file objects."
-    try:
-        for obj in open_fds:
-            if isinstance(obj, int):
-                if obj >= 0:
-                    try:
-                        os.close(obj)
-                    except OSError:
-                        pass
-            else:
-                obj.close()
-    finally:
-        open_fds.clear()
+    return run.result()
